
"""
Movie recommender system as a showcase project
"""

import argparse
import collections
import json
import os
from pprint import pprint
import random
import re
import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow_recommenders as tfrs


def load_movielense(basedir):
    """Load an return movies and ratings as dataframes"""
    ratings_filename = os.path.join(basedir, "ratings.csv")
    ratings_df = pd.read_csv(ratings_filename, dtype={
        "userId": int,
        "movieId": int,
        "rating": float,
        "timestamp": int,
    })

    movies_filename = os.path.join(basedir, "movies.csv")
    movies_df = pd.read_csv(movies_filename, dtype={
        "movieId": int,
        "title": str,
        "genres": str,
    })

    return ratings_df, movies_df


def collect_user_histories(ratings, min_rating=2.1):
    """Create a per-user rating list"""
    user_histories = collections.defaultdict(lambda: [])  # dict mapping ids to movies

    ratings = ratings.sort_values(by=['userId', 'timestamp'])

    for row in ratings.itertuples():
        user_id = row.userId
        movie_id = row.movieId
        rating = row.rating

        if rating >= min_rating:
            user_histories[user_id].append(movie_id)

    return user_histories


def create_single_user_examples(user_history, min_len=3, max_len=10):
    """Create examples by sliding a 3-100 window over the history"""
    examples = []
    for label_idx in range(min_len, len(user_history)):
        start_idx = max(0, label_idx - max_len)
        context_movie_ids = user_history[start_idx:label_idx]
        if len(context_movie_ids) < min_len:
            continue

        while len(context_movie_ids) < max_len:
            context_movie_ids.append(0)

        movie_id = user_history[label_idx]

        feature = {
            "context_movie_id": tf.train.Feature(
                int64_list=tf.train.Int64List(value=context_movie_ids)
             ),
             "label_movie_id": tf.train.Feature(
                int64_list=tf.train.Int64List(value=[movie_id])
             )
        }
        example = tf.train.Example(features=tf.train.Features(feature=feature))
        examples.append(example)

    return examples


def create_user_examples(user_histories, frac=0.8, random_seed=20220827, **kwds):
    """Create user examples for each user history and return train,test lists"""
    examples = []
    progress_bar = tf.keras.utils.Progbar(len(user_histories))
    for user_history in user_histories.values():
        single_user_examples = create_single_user_examples(
            user_history, **kwds
        )
        examples.extend(single_user_examples)
        progress_bar.add(1)

    random.seed(random_seed)
    random.shuffle(examples)

    last_train_index = int(len(examples) * frac)
    train_examples = examples[:last_train_index]
    test_examples = examples[last_train_index:]
    return train_examples, test_examples


def parse_title_year(movie_name):
    """Return (title, year) from full title"""
    pattern = r"^\s*(.+\S)\s+\((\d+)\)\s*$"
    rem = re.match(pattern, movie_name)
    if rem:
        return rem.group(1), int(rem.group(2))
    return movie_name, None


def index_titles(movies_df):
    """Create dict index of titles"""
    titles = {}
    for row in movies_df.itertuples():
        movie_id = row.movieId
        title = row.title

        titles[movie_id] = parse_title_year(title)
    return titles


def write_user_examples(examples, filename):
    """Write examples to rt Record"""
    progress_bar = tf.keras.utils.Progbar(len(examples))
    with tf.io.TFRecordWriter(filename) as record_writer:
        for example in examples:
            record_writer.write(example.SerializeToString())
            progress_bar.add(1)


def write_movie_titles(titles, filename):
    """Write movie titles to json file"""
    with open(filename, "w", encoding="utf-8") as fileobj:
        json.dump(titles, fileobj)


def prepare_dataset(args):
    """Read raw CSV files and write TF Record"""
    ratings_df, movies_df = load_movielense(args.input)
    if args.debug:
        ratings_df = ratings_df[:10000]

    user_histories = collect_user_histories(ratings_df)
    train_examples, test_examples = create_user_examples(user_histories)

    train_file = os.path.join(args.output, "train.tfrecord")
    write_user_examples(train_examples, train_file)
    n_train = len(train_examples)
    print(f"File {train_file} with {n_train:d} records created.")

    test_file = os.path.join(args.output, "test.tfrecord")
    write_user_examples(test_examples, test_file)
    n_test = len(test_examples)
    print(f"File {test_file} with {n_test:d} records created.")


    movie_titles = index_titles(movies_df)
    titles_file = os.path.join(args.output, "titles.json")
    write_movie_titles(movie_titles, titles_file)
    n_titles = len(movie_titles)
    print(f"File {titles_file} with {n_titles:d} titles created.")


def load_dataset(basedir):
    """Load prepared dataset"""
    train_filename = os.path.join(basedir, "train.tfrecord")
    test_filename = os.path.join(basedir, "test.tfrecord")

    train_rec = tf.data.TFRecordDataset(train_filename)
    test_rec = tf.data.TFRecordDataset(test_filename)

    feature_description = {
        'context_movie_id': tf.io.FixedLenFeature([10], tf.int64, default_value=np.zeros(10)),
        'label_movie_id': tf.io.FixedLenFeature([1], tf.int64, default_value=0),
    }

    def _record_parser(example_proto):
        """Protobufer decode single example"""
        return tf.io.parse_single_example(example_proto, feature_description)

    train_ds = train_rec.map(_record_parser)
    test_ds = test_rec.map(_record_parser)

    titles_filename = os.path.join(basedir, "titles.json")
    with open(titles_filename, encoding="utf-8") as fileobj:
        titles = json.load(fileobj)

    titles = {int(k): v for k, v in titles.items()}

    movies_tensor = tf.constant([[x] for x in titles.keys()])
    movie_ds = tf.data.Dataset.from_tensor_slices(movies_tensor)

    # Cache and batch
    train_ds = train_ds.batch(128).cache()
    test_ds = test_ds.batch(128).cache()

    return train_ds, test_ds, titles, movie_ds


class RecommenderModel(tfrs.Model):
    """Recommender model tying query and candidate models together"""

    def __init__(self,  query_model, cand_model, retrieval_task):
        """New instance from query and candidate models and retrieval task"""
        super().__init__()
        self.query_model = query_model
        self.cand_model = cand_model

        self.retrieval_task = retrieval_task


    def compute_loss(self, inputs, training=False):
        """Compute and return the loss"""
        context_movie_id = inputs["context_movie_id"]
        embedded_query = self.query_model(context_movie_id)

        label_movie_id = inputs["label_movie_id"]
        embedded_cand = self.cand_model(label_movie_id)


        return self.retrieval_task(embedded_query, embedded_cand, compute_metrics=not training)


def build_embedding_model(movies, embedding_dimension):
    """Build and return query and candidate models"""
    vocab = np.concatenate(list(movies))
    query_model = tf.keras.Sequential([
        tf.keras.layers.IntegerLookup(vocabulary=vocab),
        tf.keras.layers.Embedding(len(movies) + 1, embedding_dimension),
        tf.keras.layers.GRU(embedding_dimension),
    ])

    cand_model = tf.keras.Sequential([
        tf.keras.layers.IntegerLookup(vocabulary=vocab, mask_token=None, input_shape=(1,)),
        tf.keras.layers.Embedding(len(movies) + 1, embedding_dimension),
        tf.keras.layers.Reshape((32,), input_shape=(1, 32))
    ])

    return query_model, cand_model


def build_model(movies,
                learning_rate=0.1,
                embedding_dimension=32):
    """Return the full keras/tensorflow model"""
    query_model, cand_model = build_embedding_model(
        movies, embedding_dimension=embedding_dimension
    )

    metrics = tfrs.metrics.FactorizedTopK(
      candidates=movies.batch(128).map(cand_model)
    )

    retrieval_task= tfrs.tasks.Retrieval(
      metrics=metrics
    )

    model = RecommenderModel(query_model, cand_model, retrieval_task)
    model.compile(optimizer=tf.keras.optimizers.Adagrad(
        learning_rate=learning_rate
    ))

    return model


def fit_model(model, training_ds, epochs=3):
    """Fit the model to the training dataset"""
    model.fit(training_ds, epochs=epochs)


def lookup(titles, movie_ids):
    """Convert list of movie ids to movie titles, years"""
    return [titles[id] for id in movie_ids]


def eval_model(model, test_ds):
    """Print the test set performance"""
    result = model.evaluate(test_ds, return_dict=True)
    pprint(result)


def train(args):
    """Load the dataset, train and evaluate the model"""
    train_ds, test_ds, _, movies = load_dataset(args.input)

    with tf.device("cpu:0"):
        model = build_model(movies)
        fit_model(model, train_ds)
        eval_model(model, test_ds)


commands = {
    "prepare": prepare_dataset,
    "train": train,
}

def get_default_parser():
    """Return the default command-line args parser"""
    parser = argparse.ArgumentParser()
    parser.add_argument("command", metavar="CMD", choices=commands,
                        help="Operation to execute")
    parser.add_argument("-i", "--input", metavar="PATH",
                        help="Path to input file")
    parser.add_argument("-o", "--output", metavar="PATH",
                        help="Path to output file(s)")
    parser.add_argument("--debug", default=False, action="store_true",
                        help="Limit test data size")

    return parser


def cli_main(args):
    """Handle CLI args and call requested command"""
    commands[args.command](args)


if __name__ == "__main__":
    parser_ = get_default_parser()
    args_ = parser_.parse_args()
    cli_main(args_)
